from django.test import TestCase
from django import forms
from wagtail import blocks
from .. import block_forms

TEST_BLOCK_DATA = [
    {
        "type": "heading_component",
        "value": {
            "heading": "This is pretty cool!!!",
            "subheading": "RIGHT?! FUCK YES!"
        },
        "id": "0bc9f67e-f116-48a6-9ca1-6c11d39b54e8"
    },
    {
        "type": "heading_component",
        "value": {
            "heading": "AWESOME!!",
            "subheading": "RIGHT?!"
        },
        "id": "d543a6bf-34dc-4365-a3fa-d302561930ae"
    },
    {
        "type": "heading_component",
        "value": {
            "heading": "WORKS NOW!",
            "subheading": "!!!!!!!!!!!!1"
        },
        "id": "c49abcae-3c66-4fc7-979d-35407226b9f5"
    },
    {
        "type": "heading_component",
        "value": {
            "heading": "Heading!!!!",
            "subheading": "Subheading"
        },
        "id": "7bd7bc3a-1d2d-4182-8726-b257beace968"
    },
    {
        "type": "heading_component",
        "value": {
            "heading": "Hey!",
            "subheading": "Subheading!"
        },
        "id": "74a94baa-acf4-49ab-be5f-9c8a70cbc623"
    },
    {
        "type": "flat_menu_component",
        "value": {
            "title": "Test Title123123! HAHA!",
            "subtitle": "<p data-block-key=\"306j3\">i am so<b><i> happy</i></b></p>",
            "items": [
                {
                    "type": "item",
                    "value": {
                        "link": {
                            "text": "Test Item 1"
                        }
                    },
                    "id": "c757f54d-0df5-4b35-8a06-4174f180ec41"
                },
                {
                    "type": "item",
                    "value": {
                        "link": {
                            "text": "Test Item 2"
                        }
                    },
                    "id": "ec3d73d1-fd01-49ba-840a-d44586ac0025"
                },
                {
                    "type": "item",
                    "value": {
                        "link": {
                            "text": "Test Item 3"
                        }
                    },
                    "id": "a98a19c6-2ead-4e69-9ea2-3158c7e82976"
                },
                {
                    "type": "item",
                    "value": {
                        "link": {
                            "text": "Test Item 4"
                        }
                    },
                    "id": "db7183a2-d9dd-4fbd-9e42-fd2b9ebf0458"
                }
            ]
        },
        "id": "3e9144fd-5fa5-47f8-917e-8fe87c15da01"
    }
]


class HeadingComponent(blocks.StructBlock):
    heading = blocks.CharBlock(max_length=25)
    subheading = blocks.CharBlock(max_length=40)

class LinkBlock(blocks.StructBlock):
    text = blocks.CharBlock(max_length=25)

class MenuItemBlock(blocks.StructBlock):
    link = LinkBlock()

class FlatMenuComponent(blocks.StructBlock):
    title = blocks.CharBlock(max_length=25)
    subtitle = blocks.RichTextBlock()
    items = blocks.ListBlock(
        MenuItemBlock()
    )

import uuid

_model_map = {}

class FakeModel(object):
    def __init__(self, title="Fake Model"):
        self.pk = uuid.uuid4()
        self.title = title
        self.attrs = {}

    def __setattr__(self, key, value):
        if key in ["pk", "title", "attrs"]:
            object.__setattr__(self, key, value)
        else:
            self.attrs[key] = value

    def __getattribute__(self, key):
        if key in ["pk", "title", "attrs", "full_clean", "save"]:
            return object.__getattribute__(self, key)
        return self.attrs.get(key)
    
    def full_clean(self):
        pass

    def save(self):
        _model_map[self.pk] = self

class FakeUser:
    def __init__(self, is_authenticated=True, is_staff=True):
        self.is_authenticated = is_authenticated
        self.is_staff = is_staff

class FakeRequest:
    def __init__(self, method="GET", user=None, GET=None, POST=None):
        self.user = user or FakeUser()
        self.method = method
        self.GET = GET
        self.POST = POST

class TestBlocks(TestCase):
    def setUp(self) -> None:
        self.stream_block = blocks.StreamBlock([
            ("heading_component", HeadingComponent()),
            ("flat_menu_component", FlatMenuComponent())
        ])
        self.stream_value = self.stream_block.to_python(TEST_BLOCK_DATA)

    def test_find_block(self):
        block, contentpath = block_forms.find_block("d543a6bf-34dc-4365-a3fa-d302561930ae", self.stream_value)
        self.assertEqual(block.value["heading"], "AWESOME!!")
        self.assertEqual(block.value["subheading"], "RIGHT?!")

        block, contentpath = block_forms.find_block("a98a19c6-2ead-4e69-9ea2-3158c7e82976", self.stream_value)
        self.assertEqual(block.value["link"]["text"], "Test Item 3")
        self.assertEqual(contentpath, ["3e9144fd-5fa5-47f8-917e-8fe87c15da01", "items", "a98a19c6-2ead-4e69-9ea2-3158c7e82976"])

        item = self.stream_block.get_block_by_content_path(self.stream_value, contentpath)
        self.assertEqual(item.value["link"]["text"], "Test Item 3")

        block, contentpath = block_forms.find_block("invalid-id", self.stream_value)
        self.assertIsNone(block)
        self.assertEqual(contentpath, [])

    def test_get_form_class(self):
        block = block_forms.find_block("d543a6bf-34dc-4365-a3fa-d302561930ae", self.stream_value)[0]
        form_class = block_forms.get_form_class(block, block.block, request=FakeRequest())
        self.assertIsNotNone(form_class)

        fields = {
            "heading": forms.CharField(max_length=25),
            "subheading": forms.CharField(max_length=40)
        }

        for name, field in form_class.base_fields.items():
            self.assertIsInstance(field, fields[name].__class__)

        VALID_DATA = {
            "heading": "New Heading",
            "subheading": "New Subheading"
        }

        INVALID_DATA = {
            "heading": "New Heading",
            "subheading": "New Subheading" * 10
        }

        valid_form = form_class(data=VALID_DATA, block=block, parent_instance=FakeModel(), request=FakeRequest(method="POST", POST=VALID_DATA))
        self.assertTrue(valid_form.is_valid())

        invalid_form = form_class(data=INVALID_DATA, block=block, parent_instance=FakeModel(), request=FakeRequest(method="POST", POST=INVALID_DATA))
        self.assertFalse(invalid_form.is_valid())

    def test_get_subblock_form_class(self):
        block = block_forms.find_block("a98a19c6-2ead-4e69-9ea2-3158c7e82976", self.stream_value)[0]
        form_class = block_forms.get_form_class(block, block.block, request=FakeRequest())
        self.assertIsNotNone(form_class)

        fields = {
            "link": block_forms.BlockEditSubFormField
        }

        subfields = {
            "text": forms.CharField(max_length=25)
        }

        for name, field in form_class.base_fields.items():
            self.assertIsInstance(field, fields[name])

        field = form_class.declared_fields["link"]
        self.assertIsInstance(field, block_forms.BlockEditSubFormField)
        self.assertIsInstance(field.form, block_forms.BaseBlockEditForm)

        for name, field in field.form.base_fields.items():
            self.assertIsInstance(field, subfields[name].__class__)

        VALID_DATA = {
            "link-text": "New Link Text"
        }

        INVALID_DATA = {
            "link-text": ["New Link Text" * 10]
        }


        valid_form = form_class(data=VALID_DATA, block=block, parent_instance=FakeModel(), request=FakeRequest(method="POST"))
        if not valid_form.is_valid():
            self.fail((
                valid_form.errors.items(),
                valid_form.non_field_errors()
            ))

        valid_form.save()
        self.assertEqual(block.value["link"]["text"], "New Link Text")

        invalid_form = form_class(data=INVALID_DATA, block=block, parent_instance=FakeModel(), request=FakeRequest(method="POST"))
        self.assertFalse(invalid_form.is_valid())


    def test_block_form_save(self):
        block = block_forms.find_block("d543a6bf-34dc-4365-a3fa-d302561930ae", self.stream_value)[0]
        form_class = block_forms.get_form_class(block, block.block, request=FakeRequest())

        VALID_DATA = {
            "heading": "New Heading",
            "subheading": "New Subheading"
        }

        valid_form = form_class(data=VALID_DATA, block=block, parent_instance=FakeModel(), request=FakeRequest(method="POST"))
        if not valid_form.is_valid():
            self.fail((valid_form.errors.items(), valid_form.non_field_errors()))

        valid_form.save()
        self.assertEqual(block.value["heading"], "New Heading")
        self.assertEqual(block.value["subheading"], "New Subheading")


    def test_subblock_form_save(self):
        block = block_forms.find_block("a98a19c6-2ead-4e69-9ea2-3158c7e82976", self.stream_value)[0]
        form_class = block_forms.get_form_class(block, block.block, request=FakeRequest())

        VALID_DATA = {
            "link-text": "New Link Text"
        }

        valid_form = form_class(data=VALID_DATA, block=block, parent_instance=FakeModel(), request=FakeRequest(method="POST"))
        if not valid_form.is_valid():
            self.fail((
                valid_form.errors.items(),
                valid_form.non_field_errors()
            ))

        valid_form.save()
        self.assertEqual(block.value["link"]["text"], "New Link Text")



from typing import Any, Union
from django.views.generic import TemplateView
from django.utils.translation import gettext_lazy as _
from django.shortcuts import redirect
from django.contrib import messages
from django.urls import reverse
from django.apps import apps
from django.http import (
    HttpRequest,
    HttpResponseBadRequest,
    HttpResponse,
)

from wagtail.admin.views.generic import WagtailAdminTemplateMixin
from wagtail.models import (
    RevisionMixin,
    PreviewableMixin,
    DraftStateMixin,
)
from .. import block_forms as block_forms
from ..utils import (
    FEDIT_PREVIEW_VAR,
    USERBAR_MODEL_VAR,
    FeditPermissionCheck,
    with_userbar_model,
)


class BaseFeditView(FeditPermissionCheck, TemplateView):
    def dispatch(self, request: HttpRequest, object_id: Any, app_label: str, model_name: str) -> HttpResponse:

        self.model = apps.get_model(app_label, model_name)
        self.model_object = self.model._default_manager.get(pk=object_id)

        if not self.has_perms(request, self.model):
            return HttpResponseBadRequest("You do not have permission to view this page")

        if issubclass(self.model, (RevisionMixin, PreviewableMixin)):
            instance: RevisionMixin  = self.model_object
            revision: Union[PreviewableMixin, RevisionMixin] = instance.latest_revision
            self.object = revision.as_object()
            self.is_preview = True
        else:
            self.object = self.model_object
            self.is_preview = False

        try:
            self.checks(request, self.object)
        except ValueError as e:
            return HttpResponseBadRequest(str(e))

        return super().dispatch(request, object_id, app_label, model_name)
    
    def checks(self, request: HttpRequest, object: Any) -> None:
        pass

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        return super().get_context_data(**kwargs) | {
            "object": self.object,
            "model": self.model,
        }

class FEditableView(BaseFeditView):

    def checks(self, request: HttpRequest, object: Any) -> None:
        super().checks(request, object)
        if not isinstance(self.object, PreviewableMixin):
            raise ValueError("Model {} does not inherit from PreviewableMixin, cannot edit.".format(self.model))

    def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        self.request = with_userbar_model(self.request, self.object)
        object: PreviewableMixin = self.object
        return object.make_preview_request(original_request=self.request, extra_request_attrs={
            FEDIT_PREVIEW_VAR: True,
            USERBAR_MODEL_VAR: self.object,
        })
        

class FeditablePublishView(WagtailAdminTemplateMixin, BaseFeditView):
    template_name = "wagtail_fedit/editor/publish_confirm.html"
    object: DraftStateMixin

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        return super().get_context_data(**kwargs) | {
            "model_verbose_name": self.object._meta.verbose_name,
            "object": self.object,
            "publish_url": reverse(
                "wagtail_fedit:publish",
                args=[self.object.pk, self.model._meta.app_label, self.model._meta.model_name],
            ),
            "edit_url": reverse(
                "wagtail_fedit:editable",
                args=[self.object.pk, self.model._meta.app_label, self.model._meta.model_name],
            ),
        }
    
    def get_header_title(self):
        return _("Publishing {} \"{}\"").format(self.object._meta.verbose_name, self.object)

    def checks(self, request: HttpRequest, object: Any) -> None:
        if not isinstance(self.object, DraftStateMixin):
            raise ValueError("Model {} does not inherit from DraftStateMixin, cannot publish.".format(self.model))
    
    def post(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:

        if "publish" not in request.POST:
            return HttpResponseBadRequest("You must confirm the publish action")
        
        latest_revision = None
        if isinstance(self.object, RevisionMixin):
            latest_revision = self.object.latest_revision

        self.object.publish(
            revision=latest_revision,
            user=request.user,
            skip_permission_checks=False,
        )

        if latest_revision:
            self.object = latest_revision.as_object()

        # messages.success(
        #     request,
        #     _("Published {} \"{}\"").format(
        #         self.object._meta.verbose_name,
        #         self.object
        #     )
        # )
            
        if hasattr(self.object, "get_url"):
            return redirect(self.object.get_url(request))
        
        elif hasattr(self.object, "url"):
            return redirect(self.object.url)
        
        elif hasattr(self.object, "get_absolute_url"):
            return redirect(self.object.get_absolute_url())
        
        return redirect(reverse(
            "wagtail_fedit:editable",
            args=[self.object.pk, self.model._meta.app_label, self.model._meta.model_name],
        ))



from typing import Any
from django.http import HttpRequest
from django.db import models
from wagtail import hooks

from .hooks import EXCLUDE_FROM_RELATED_FORMS


FEDIT_PREVIEW_VAR = "_wagtail_fedit_preview"
USERBAR_MODEL_VAR = "_wagtail_fedit_userbar_model"


class FeditPermissionCheck:
    @staticmethod
    def has_perms(request: HttpRequest, model: Any) -> bool:
        
        if (
            not request.user.is_authenticated\
            or not request.user.has_perm("wagtailadmin.access_admin")\
            or not request.user.has_perm(f"{model._meta.app_label}.change_{model._meta.model_name}")    
        ):
            return False
        
        return True


def use_related_form(field: models.Field) -> bool:
    for hook in hooks.get_hooks(EXCLUDE_FROM_RELATED_FORMS):
        if hook(field):
            return False
    return True


def access_userbar_model(request: HttpRequest) -> models.Model:

    if not hasattr(request, USERBAR_MODEL_VAR):
        return None
    
    return getattr(request, USERBAR_MODEL_VAR)

def with_userbar_model(request: HttpRequest, model: models.Model) -> HttpRequest:
    setattr(request, USERBAR_MODEL_VAR, model)
    return request



from django import forms
from django.http import (
    HttpRequest,
)

from wagtail import blocks
from wagtail import hooks

from globlocks.blocks import BaseBlockConfiguration
import inspect

from ..hooks import REGISTER_BLOCK_HOOK_NAME
from .subform_field import BlockEditSubFormField



block_map = {
    blocks.CharBlock: (forms.CharField, forms.TextInput),
    blocks.TextBlock: (forms.CharField, forms.Textarea),
    blocks.IntegerBlock: forms.IntegerField,
    blocks.FloatBlock: forms.FloatField,
    blocks.DecimalBlock: forms.DecimalField,
    blocks.BooleanBlock: forms.BooleanField,
    blocks.DateBlock: forms.DateField,
    blocks.TimeBlock: forms.TimeField,
    blocks.DateTimeBlock: forms.DateTimeField,
    blocks.URLBlock: forms.URLField,
    blocks.EmailBlock: forms.EmailField,
    blocks.ChoiceBlock: lambda block, request: forms.ChoiceField(choices=block._constructor_kwargs["choices"]),
    blocks.MultipleChoiceBlock: lambda block, request: forms.MultipleChoiceField(choices=block._constructor_kwargs["choices"]),
    blocks.RawHTMLBlock: (forms.CharField, forms.Textarea),
    blocks.RichTextBlock: lambda block, request: block.field,
}

_looked_for_conversions = False

def look_for_conversions():
    global _looked_for_conversions
    if _looked_for_conversions:
        return

    _looked_for_conversions = True

    for hook in hooks.get_hooks(REGISTER_BLOCK_HOOK_NAME):
        hook(block_map)



from .field_conversions import (
    block_map, look_for_conversions
)
from .subform_field import (
    BlockEditSubFormField, SubFormWidget,
)
from .forms import (
    get_field_for_block,
    get_form_class,
    BaseBlockEditForm,
)
from .utils import (
    find_block,
    get_block_name,
    get_initial_for_form,
)


from django import forms
from django.http import (
    HttpRequest,
)
from django.db import models
from wagtail import blocks
from wagtail.models import RevisionMixin
import inspect

from . import utils
from .subform_field import (
    BlockEditSubFormField,
    SubformValidationError,
)
from .field_conversions import (
    block_map, look_for_conversions,
)


def get_field_for_block(field_name, parent_instance: blocks.BoundBlock, block: blocks.Block, request: HttpRequest):

    if not utils.can_fedit(block):
        # Marked as not editable on the frontend.
        return None

    elif type(block) in block_map:
        widget = None
        field = block_map[type(block)]
        
        if isinstance(field, tuple):
            field, widget = field
        
        if isinstance(field, dict):
            field = field["func"](block, request)

        if inspect.isfunction(field):
            field = field(block, request)

        if type(field) == type:
            field = field(
                widget=widget,
                required=getattr(block.meta, "required", False),
                label=getattr(block.meta, "label", None),
                help_text=getattr(block.meta, "help_text", None),
            )

    elif isinstance(block, blocks.StructBlock):

        form_class = get_form_class(
            parent_instance.value.bound_blocks[field_name],
            block,
            request,
            base_class=BaseBlockEditForm
        )

        if not form_class:
            return None

        field = BlockEditSubFormField(
            parent_instance.value.bound_blocks[field_name],
            form_class,
            request,
        )

    elif hasattr(block, "field"):
        field = block.field

    else:
        return None
        # raise ValueError("Invalid block type: %s" % type(block))
    
    return field



def get_form_class(instance: blocks.BoundBlock, block: blocks.Block, request, base_class=None):
    look_for_conversions()

    if not base_class:
        base_class = BlockEditForm

    if hasattr(block, "fedit_form_class"):
        return block.fedit_form_class()

    if isinstance(block, blocks.StructBlock):
        fields = []
        for field_name, sub in block.child_blocks.items():

            #if isinstance(sub, (blocks.StructBlock, blocks.StreamBlock, blocks.ListBlock, blocks.RichTextBlock)):
            #    continue
            field = get_field_for_block(field_name, instance, sub, request)
            if not field:
                continue

            field.help_text = getattr(sub.meta, "help_text", field.help_text)

            fields.append(
                (field_name, field)
            )

        return type(
            block.__class__.__name__ + "Form",
            (base_class,),
            dict(fields),
        )
    
    field = get_field_for_block("value", instance, block, request)
    if not field:
        return None
    
    field.label = block.label
    field.help_text = getattr(block.meta, "help_text", field.help_text)

    return type(
        block.__class__.__name__ + "Form",
        (base_class,),
        {"value": field},
    )



class BaseBlockEditForm(forms.Form):
    def __init__(self, *args, block: blocks.BoundBlock, request: HttpRequest, **kwargs):
        self.block = block
        self.request = request
        self._subform_errors = {}
        kwargs["initial"] = utils.get_initial_for_form(block, block)
        super().__init__(*args, **kwargs)

    # def add_error(self, field: str | None, error: forms.ValidationError | str) -> None:
    #     if isinstance(error, SubformValidationError):
    #         self._subform_errors[field] = error.subform_errors
    #     else:
    #         super().add_error(field, error)
# 
    # def has_error(self, field: str, code: str | None = ...) -> bool:
    #     return super().has_error(field, code) or field in self._subform_errors

    def clean(self):
        cleaned_data = super().clean()

        if hasattr(self.block, "fedit_clean"):
            self.block.fedit_clean(cleaned_data, self.parent_instance)
            return cleaned_data
    
        if isinstance(self.block, blocks.BoundBlock):
            block = self.block.block
            if hasattr(block, "fedit_clean"):
                block.fedit_clean(cleaned_data, self.parent_instance)
                return cleaned_data
            
            if isinstance(self.block.value, blocks.StructValue):
                for field_name, sub in block.child_blocks.items():
                    if field_name not in cleaned_data:
                        continue

                    value = cleaned_data.get(field_name)
                    if isinstance(value, blocks.BoundBlock):
                        value = value.value
                    try:
                        if hasattr(sub, "field"):
                            value = sub.field.clean(value)
                        else:
                            value = sub.clean(value)
                        
                        if isinstance(value, blocks.BoundBlock):
                            value = value.value

                        cleaned_data[field_name] = value

                    except blocks.StructBlockValidationError as e:
                        sub = block.child_blocks[field_name]
                        for name, b in sub.child_blocks.items():
                            if name in e.block_errors:
                                errors = e.block_errors[name]
                                del e.block_errors[name]
                                e.block_errors[b.label] = errors

                        for name, errors in e.block_errors.items():
                            self.add_error(field_name, f"{name}: {', '.join(errors)}")

            else:
                self.cleaned_data["value"] = block.clean(cleaned_data["value"])

        return cleaned_data
    
    @property
    def errors(self):
        errors = super().errors
        errors.update(self._subform_errors)
        return errors

    def save(self):
        block = self.block

        if hasattr(block, "fedit_save"):
            block.fedit_save(self.cleaned_data, self.parent_instance)
        else:

            if isinstance(block, blocks.BoundBlock):

                if hasattr(block.block, "fedit_save"):
                    block.block.fedit_save(self.cleaned_data, self.parent_instance)


                if isinstance(block.value, blocks.StructValue):
                    for field_name, value in self.cleaned_data.items():
                        block.value[field_name] = utils.value_for_form(
                            block.value[field_name],
                            block.block.child_blocks[field_name],
                            value
                        )

                else:
                    value = self.cleaned_data["value"]
                    # fieldblock special method - great!
                    block.value = block.block.value_from_form(value)
            else:
                raise ValueError("Invalid block type: %s" % type(block))

        return block

class BlockEditForm(BaseBlockEditForm):
    def __init__(self, *args, parent_instance: models.Model, **kwargs):
        super().__init__(*args, **kwargs)
        self.parent_instance = parent_instance

    def save(self):
        block = super().save()
        self.parent_instance.full_clean()

        if isinstance(self.parent_instance, RevisionMixin):
            self.parent_instance = self.parent_instance.save_revision(
                user=self.request.user,
            )
        else:
            self.parent_instance = self.parent_instance.save()
        return block
        


from django.urls import path, include
from . import views

app_name = "wagtail_fedit"

urlpatterns = [
    path("editable/<str:object_id>/<str:app_label>/<str:model_name>/", views.FEditableView.as_view(), name="editable"),
    path("publish/<str:object_id>/<str:app_label>/<str:model_name>/", views.FeditablePublishView.as_view(), name="publish"),
    path("field/<str:field_name>/<str:app_label>/<str:model_name>/<str:model_id>/", views.EditFieldView.as_view(), name="edit_field"),
    path("edit_block/<str:block_id>/<str:field_name>/<str:app_label>/<str:model_name>/<str:model_id>/", views.EditBlockView.as_view(), name="edit_block"),
]


from typing import Any, Type
from django import forms
from django.http import (
    HttpRequest,
)
from wagtail import blocks
from .utils import get_block_name

class SubformValidationError(forms.ValidationError):
    def __init__(self, message, code=None, params=None):
        if isinstance(message, dict):
            self.subform_errors = message
            message = list(message.values())
        else:
            self.subform_errors = {}
        super().__init__(message, code, params)

    def __repr__(self):
        return str(self.subform_errors)


class SubFormWidget(forms.Widget):
    def __init__(self,
            block:   blocks.BoundBlock,
            form:    forms.Form,
            request: HttpRequest,
            attrs=None,
        ):
        self.block = block
        self.form = form
        self.request = request
        super().__init__(attrs)

    def render(self, name, value, attrs=None, renderer=None):
        return self.form.as_p()
    
    def value_from_datadict(self, data, files, name):
        return self.form.__class__(data=data, files=files, prefix=name, block=self.block, request=self.request)

    def format_value(self, value: Any):
        return value

    @property
    def media(self):
        return self.form.media
    

class BlockEditSubFormField(forms.Field):
    def __init__(self,
            block: blocks.BoundBlock,
            form: Type[forms.Form],
            request: HttpRequest, 
            *args,
            **kwargs
        ):
        self.block:   blocks.BoundBlock = block
        self.request: HttpRequest = request

        if self.request.method == "POST":
            self.form = form(prefix=self.block_name, data=request.POST, block=block, request=request)
        else:
            self.form = form(prefix=self.block_name, block=block, request=request)

        super().__init__(*args, **kwargs)
        self.widget = SubFormWidget(block, self.form, request)

    @property
    def block_name(self):
        return get_block_name(self.block)

    def validate(self, value: Any):
        super().validate(value)

        if not value.is_valid():
            raise SubformValidationError(value.errors)

    def clean(self, value: Any) -> Any:
        value = super().clean(value)
        return value.save()



from wagtail import blocks
from wagtail.blocks.list_block import ListValue
from wagtail.fields import StreamValue

def get_block_name(block):
    if isinstance(block, StreamValue.StreamChild):
        return block.block_type
    elif isinstance(block, ListValue.ListChild):
        return "item"
    elif isinstance(block, ListValue):
        return block.block.name
    elif isinstance(block, (blocks.StructValue, blocks.BoundBlock)):
        return block.block.name
    else:
        raise ValueError("Unknown block type: %s" % type(block))

def get_block_path(block):
    if isinstance(block, StreamValue.StreamChild):
        return block.id
    elif isinstance(block, ListValue.ListChild):
        return block.id
    elif isinstance(block, ListValue):
        return block.block.name
    elif isinstance(block, (blocks.StructValue, blocks.BoundBlock)):
        return block.block.name
    else:
        raise ValueError("Unknown block type: %s" % type(block))


def find_block(block_id, field, contentpath=None):
    if contentpath is None:
        contentpath = []

    # Check and cast field to iterable if necessary, but do not append non-StreamValue names to contentpath here.
    if not isinstance(field, StreamValue) and not hasattr(field, "__iter__"):
        field = [field]

    # Adjust for ListValue to get the iterable bound_blocks.
    if isinstance(field, ListValue):
        field = field.bound_blocks

    for block in field:
        # Determine the block's name only if needed to avoid premature addition to contentpath.
        block_name = get_block_path(block)
        
        if getattr(block, "id", None) == block_id:
            # Append the block name here as it directly leads to the target.
            return block, contentpath + [block_name]
        
        # Prepare to check children without altering the current path yet.
        if isinstance(block.value, blocks.StructValue):
            for _, value in block.value.bound_blocks.items():
                found, found_path = find_block(block_id, value, contentpath + [block_name])
                if found:
                    return found, found_path

        elif isinstance(block.value, (StreamValue, StreamValue.StreamChild, ListValue)):
            found, found_path = find_block(block_id, block.value, contentpath + [block_name])
            if found:
                return found, found_path

    # Return None and the current path if no block is found at this level.
    return None, contentpath


def can_fedit(block):
    """
        Check if a block is marked as non-editable on the frontend.
    """
    if isinstance(block, blocks.BoundBlock):
        return getattr(block.block.meta, "fedit", True)

    return getattr(block.meta, "fedit", True)


def value_for_form(block_value, block, value):
    if not block:
        return value
    
    if isinstance(value, blocks.BoundBlock):
        return value.value
    
    if hasattr(block, "fedit_value_for_form"):
        return block.fedit_value_for_form(value)
    
    if isinstance(block, blocks.StructBlock):
        for field_name, sub in block.child_blocks.items():
            if not value:
                continue
            block_value[field_name] = value_for_form(block_value[field_name], sub, value[field_name])
        return block_value
    
    if isinstance(block, blocks.FieldBlock):
        return block.value_from_form(value)
    
    return value


def get_initial_for_form(block_value, block, parent_structblock = False):
    if isinstance(block, (StreamValue.StreamChild, ListValue.ListChild, blocks.BoundBlock)):
        return get_initial_for_form(block.value, block.block)

    if hasattr(block, "fedit_initial_for_form"):
        return block.fedit_initial_for_form(block_value)

    if isinstance(block, (blocks.FieldBlock)) and not parent_structblock:
        return {"value": block_value}
    
    if isinstance(block, (blocks.StructBlock)):
        d = {}
        for key, value in block_value.items():
            if value and hasattr(value, "block"):
                d[key] = get_initial_for_form(value, value.block, parent_structblock=True)
            elif value:
                d[key] = value
            else:
                d[key] = None
        return d
    
    return block_value
